module cmdline_options

  use iso_varying_string, string_t => varying_string
  use diagnostics

  implicit none
  private

  public :: no_option_value
  public :: get_option_value

  integer, parameter :: CMDLINE_ARG_LEN = 1000

  abstract interface
     subroutine msg
     end subroutine msg
  end interface

  procedure (msg), pointer :: print_usage => null ()

contains

  subroutine no_option_value (option, value)
    type(string_t), intent(in) :: option, value
    if (value /= "") then
       call msg_error (" Option '" // char (option) // "' should have no value")
    end if
  end subroutine no_option_value

  function get_option_value (i, option, value) result (string)
    type(string_t) :: string
    integer, intent(inout) :: i
    type(string_t), intent(in) :: option
    type(string_t), intent(in), optional :: value
    character(CMDLINE_ARG_LEN) :: arg_value
    integer :: arg_len, arg_status
    logical :: has_value
    if (present (value)) then
       has_value = value /= ""
    else
       has_value = .false.
    end if
    if (has_value) then
       call unquote_value (i, option, value, string)
    else
       i = i + 1
       call get_command_argument (i, arg_value, arg_len, arg_status)
       select case (arg_status)
       case (0)
       case (-1)
          call msg_error (" Option value truncated: '" // arg_value // "'")
       case default
          call print_usage ()
          call msg_fatal (" Option '" // char (option) // "' needs a value")
       end select
       select case (arg_value(1:1))
       case ("-")
          call print_usage ()
          call msg_fatal (" Option '" // char (option) // "' needs a value")
       end select
       call unquote_value (i, option, var_str (trim (arg_value)), string)
    end if
  end function get_option_value

  subroutine unquote_value (i, option, value, string)
    integer, intent(inout) :: i
    type(string_t), intent(in) :: option
    type(string_t), intent(in) :: value
    type(string_t), intent(out) :: string
    character(1) :: quote
    character(CMDLINE_ARG_LEN) :: arg_value
    integer :: arg_len, arg_status
    quote = extract (value, 1, 1)
    select case (quote)
    case ("'", '"')
       string = ""
       arg_value = extract (value, 2)
       arg_len = len_trim (value)
       APPEND_QUOTED: do
          if (extract (arg_value, arg_len, arg_len) == quote) then
             string = string // " " // extract (arg_value, 1, arg_len-1)
             exit APPEND_QUOTED
          else
             string = string // " " // trim (arg_value)
             i = i + 1
             call get_command_argument (i, arg_value, arg_len, arg_status)
             select case (arg_status)
             case (0)
             case (-1)
                call msg_error (" Quoted option value truncated: '" &
                     // char (string) // "'")
             case default
                call print_usage ()
                call msg_fatal (" Option '" // char (option) &
                     // "': unterminated quoted value")
             end select
          end if
       end do APPEND_QUOTED
    case default
       string = value
    end select
  end subroutine unquote_value

end module cmdline_options

