program main_ut

  use iso_varying_string, string_t => varying_string
  use unit_tests
  use io_units
  use system_dependencies
  use diagnostics
  use os_interface

  use cmdline_options

  use model_testbed !NODEP!


  use eio_base_ut, only: eio_prepare_test
  use eio_base_ut, only: eio_cleanup_test
  use eio_base_ut, only: eio_prepare_fallback_model
  use eio_base_ut, only: eio_cleanup_fallback_model
  use dispatch_rng, only: dispatch_rng_factory_fallback
  use dispatch_rng_ut, only: dispatch_rng_factory_test
  use os_interface_ut, only: os_interface_test
  !!! This has a whitespace failure
  use parser_ut, only: parse_test
  !!! Output "-0.000"
  use bloch_vectors_ut, only: bloch_vectors_test
  !!! Wrong test (pointer issue?)
  use prclib_stacks_ut, only: prclib_stacks_test
  use rt_data_ut, only: rt_data_test !!! (rt_data_3)
  !!! Seg  faults !!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  use sf_base_ut, only: sf_base_test
  use sf_isr_ut, only: sf_isr_test
  use phs_base_ut, only: phs_base_test
  use jets_ut, only: jets_test
  use particle_specifiers_ut, only: particle_specifiers_test
  use phs_forests_ut, only: phs_forests_test
  use phs_wood_ut, only: phs_wood_test
  use prc_test_ut, only: prc_test_test
  use prc_template_me_ut, only: prc_template_me_test
  use prclib_interfaces_ut, only: prclib_interfaces_test
  use process_libraries_ut, only: process_libraries_test
  use expr_tests_ut, only: subevt_expr_test
  use processes_ut, only: processes_test
  use process_stacks_ut, only: process_stacks_test
  use event_transforms_ut, only: event_transforms_test
  use resonance_insertion_ut, only: resonance_insertion_test
  use isr_epa_handler_ut, only: isr_handler_test
  use isr_epa_handler_ut, only: epa_handler_test
  use decays_ut, only: decays_test
  use events_ut, only: events_test
  use eio_base_ut, only: eio_base_test
  use eio_direct_ut, only: eio_direct_test
  use eio_raw_ut, only: eio_raw_test
  use eio_checkpoints_ut, only: eio_checkpoints_test
  use eio_lhef_ut, only: eio_lhef_test
  use eio_ascii_ut, only: eio_ascii_test
  use eio_weights_ut, only: eio_weights_test
  use eio_dump_ut, only: eio_dump_test
  use dispatch_rng_ut, only: dispatch_rng_test
  use process_configurations_ut, only: process_configurations_test
  use compilations_ut, only: compilations_test
  use compilations_ut, only: compilations_static_test
  use integrations_ut, only: integrations_test
  use integrations_ut, only: integrations_history_test
  use event_streams_ut, only: event_streams_test
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  implicit none

  integer, parameter :: CMDLINE_ARG_LEN = 1000
  character(CMDLINE_ARG_LEN) :: arg
  character(2) :: option
  type(string_t) :: long_option, value
  integer :: i, j, arg_len, arg_status
  logical :: look_for_options
  type(string_t) :: check, checks
  type(test_results_t) :: test_results
  logical :: success
  integer :: quit_code = 0
  look_for_options = .true.
  logging = .false.
  msg_level = RESULT
  check = ""
  checks = ""

  i = 0
  SCAN_CMDLINE: do
     i = i + 1
     call get_command_argument (i, arg, arg_len, arg_status)
     select case (arg_status)
     case (0)
     case (-1)
        call msg_error (" Command argument truncated: '" // arg // "'")
     case default
        exit SCAN_CMDLINE
     end select
     if (look_for_options) then
        select case (arg(1:2))
        case ("--")
           value = trim (arg)
           call split (value, long_option, "=")
           select case (char (long_option))
           case ("--version")
              call no_option_value (long_option, value)
           case ("--help")
              call no_option_value (long_option, value)
           case ("--check")
              check = get_option_value (i, long_option, value)
              checks = checks // " " // check
              cycle SCAN_CMDLINE
           case ("--debug")
              call no_option_value (long_option, value)
              call set_debug_levels (get_option_value (i, long_option, value))
              cycle SCAN_CMDLINE
           case ("--debug2")
              call no_option_value (long_option, value)
              call set_debug2_levels (get_option_value (i, long_option, value))
              cycle SCAN_CMDLINE
           case default
              call msg_fatal ("Option '" // trim (arg) // "' not recognized")
           end select
        end select
        select case (arg(1:1))
        case ("-")
           j = 1
           if (len_trim (arg) == 1) then
              look_for_options = .false.
           else
              SCAN_SHORT_OPTIONS: do
                 j = j + 1
                 if (j > len_trim (arg)) exit SCAN_SHORT_OPTIONS
                 option = "-" // arg(j:j)
                 select case (option)
                 case ("-?", "-h")
                 case default
                    call msg_fatal &
                         ("Option '" // option // "' not recognized")
                 end select
              end do SCAN_SHORT_OPTIONS
           end if
        case default
           call msg_fatal ("Option '" // trim (arg) // "' not recognized")
        end select
     else
        call msg_fatal ("Option '" // trim (arg) // "' not recognized")
     end if
  end do SCAN_CMDLINE

   if (checks /= "") then
      checks = trim (adjustl (checks))
      RUN_CHECKS: do while (checks /= "")
         call split (checks, check, " ")
         call whizard_check (check, test_results)
      end do RUN_CHECKS
      call test_results%wrapup (6, success)
      if (.not. success)  quit_code = 7
   end if

   call msg_terminate (quit_code = quit_code)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
contains

  subroutine prepare_eio_test (event, unweighted, n_alt, sample_norm)
    use variables, only: var_list_t
    use model_data
    use process, only: process_t
    use instances, only: process_instance_t
    use processes_ut, only: prepare_test_process
    use event_base
    use events

    class(generic_event_t), intent(inout), pointer :: event
    logical, intent(in), optional :: unweighted
    integer, intent(in), optional :: n_alt
    type(string_t), intent(in), optional :: sample_norm
    type(model_data_t), pointer :: model
    type(var_list_t) :: var_list
    type(string_t) :: sample_normalization
    type(process_t), pointer :: proc
    type(process_instance_t), pointer :: process_instance

    allocate (model)
    call model%init_test ()

    allocate (proc)
    allocate (process_instance)

    call prepare_test_process (proc, process_instance, model, &
         run_id = var_str ("run_test"))
    call process_instance%setup_event_data ()

    call model%final ()
    deallocate (model)

    allocate (event_t :: event)
    select type (event)
    type is (event_t)
       if (present (unweighted)) then
          call var_list%append_log (&
               var_str ("?unweighted"), unweighted, &
               intrinsic = .true.)
       else
          call var_list%append_log (&
               var_str ("?unweighted"), .true., &
               intrinsic = .true.)
       end if
       if (present (sample_norm)) then
          sample_normalization = sample_norm
       else
          sample_normalization = var_str ("auto")
       end if
       call var_list%append_string (&
            var_str ("$sample_normalization"), &
            sample_normalization, intrinsic = .true.)
       call event%basic_init (var_list, n_alt)
       call event%connect (process_instance, proc%get_model_ptr ())
       call var_list%final ()
    end select

  end subroutine prepare_eio_test

  subroutine cleanup_eio_test (event)
    use model_data
    use process, only: process_t
    use instances, only: process_instance_t
    use processes_ut, only: cleanup_test_process
    use event_base
    use events

    class(generic_event_t), intent(inout), pointer :: event
    type(process_t), pointer :: proc
    type(process_instance_t), pointer :: process_instance

    select type (event)
    type is (event_t)
       proc => event%get_process_ptr ()
       process_instance => event%get_process_instance_ptr ()
       call cleanup_test_process (proc, process_instance)
       deallocate (process_instance)
       deallocate (proc)
       call event%final ()
    end select
    deallocate (event)

  end subroutine cleanup_eio_test

  subroutine prepare_whizard_model (model, name, vars)
    use iso_varying_string, string_t => varying_string
    use os_interface
    use model_data
    use var_base
    use models
    class(model_data_t), intent(inout), pointer :: model
    type(string_t), intent(in) :: name
    class(vars_t), pointer, intent(out), optional :: vars
    type(os_data_t) :: os_data
    call syntax_model_file_init ()
    call os_data%init ()
    if (.not. associated (model))  allocate (model_t :: model)
    select type (model)
    type is (model_t)
       call model%read (name // ".mdl", os_data)
       if (present (vars)) then
          vars => model%get_var_list_ptr ()
       end if
    end select
  end subroutine prepare_whizard_model

  subroutine cleanup_whizard_model (model)
    use model_data
    use models
    class(model_data_t), intent(inout), target :: model
    call model%final ()
    call syntax_model_file_final ()
  end subroutine cleanup_whizard_model

  subroutine prepare_fallback_model (model)
    use model_data
    class(model_data_t), intent(inout), pointer :: model
    call prepare_whizard_model (model, var_str ("SM_hadrons"))
  end subroutine prepare_fallback_model

  subroutine whizard_check (check, results)
    type(string_t), intent(in) :: check
    type(test_results_t), intent(inout) :: results
    type(os_data_t) :: os_data
    integer :: u
    call os_data%init ()
    u = free_unit ()
    open (u, file="whizard_check." // char (check) // ".log", &
         action="write", status="replace")
    call msg_message (repeat ('=', 76), 0)
    call msg_message ("Running self-test: " // char (check), 0)
    call msg_message (repeat ('-', 76), 0)
    eio_prepare_test => prepare_eio_test
    eio_cleanup_test => cleanup_eio_test
    prepare_model => prepare_whizard_model
    cleanup_model => cleanup_whizard_model
    eio_prepare_fallback_model => prepare_fallback_model
    eio_cleanup_fallback_model => cleanup_model
    dispatch_rng_factory_fallback => dispatch_rng_factory_test
    select case (char (check))
    case ("test01")
       call os_interface_test (u, results)
    case ("test02")
       call parse_test (u, results)
    case ("test03")
       call phs_forests_test (u, results)
    case ("test04")
       call bloch_vectors_test (u, results)
    case ("test05")
       call sf_base_test (u, results)
    case ("test06")       
       call sf_isr_test (u, results)
    case ("test07")       
       call phs_base_test (u, results)
    case ("test08")       
       call phs_wood_test (u, results)
    case ("test10")       
       call prclib_interfaces_test (u, results)
    case ("test11")       
       call particle_specifiers_test (u, results)
    case ("test12")       
       call process_libraries_test (u, results)
    case ("test13")      
       call prclib_stacks_test (u, results)
    case ("test14")      
       call jets_test (u, results)
    case ("test15")       
       call prc_test_test (u, results)
    case ("test16")       
       call prc_template_me_test (u, results)
    case ("test19")
       call subevt_expr_test (u, results)
    case ("test20")
       call processes_test (u, results)
    case ("test21")
       call process_stacks_test (u, results)
    case ("test22")
       call event_transforms_test (u, results)
    case ("test23")
       call resonance_insertion_test (u, results)
    case ("test24")
       call isr_handler_test (u, results)
    case ("test25")
       call epa_handler_test (u, results)
    case ("test26")
       call decays_test (u, results)
    case ("test27")
       call events_test (u, results)
    case ("test28")
       call eio_base_test (u, results)
    case ("test29")
       call eio_direct_test (u, results)
    case ("test30")
       call eio_raw_test (u, results)
    case ("test31")
       call eio_checkpoints_test (u, results)
    case ("test32")
       call eio_lhef_test (u, results)
    case ("test34")
       call eio_ascii_test (u, results)
    case ("test35")
       call eio_weights_test (u, results)
    case ("test36")
       call eio_dump_test (u, results)
    case ("test37")
       call rt_data_test (u, results)
    case ("test38")
       call dispatch_rng_test (u, results)
    case ("test39")
       call process_configurations_test (u, results)
    case ("test40")
       call compilations_test (u, results)
    case ("test41")
       call compilations_static_test (u, results)
    case ("test42")
       call integrations_test (u, results)
    case ("test43")
       call integrations_history_test (u, results)
    case ("test44")
       call event_streams_test (u, results)
    case default
       call msg_fatal ("Self-test '" // char (check) // "' not implemented.")
    end select
    close (u)
  end subroutine whizard_check


end program main_ut
